import tarfile
import itertools
import math
from Bio import SeqIO
from pankmer.index import print_err, kmer_size, generate_mem_blocks
from os import listdir
from os.path import join, exists, isfile, isdir

from pankmer.version import __version__
from pankmer.gzip_agnostic_open import gzip_agnostic_open


def dryrun(genomes_input, split_memory: int = 1, threads: int = 1) -> dict:
    if not isinstance(genomes_input, (tuple, list)):
        genomes_input = list(genomes_input) if isinstance(genomes_input, set) else [genomes_input]
    if len(genomes_input) == 1 and isfile(genomes_input[0]) and tarfile.is_tarfile(genomes_input[0]):
        input_is_tar = True
        with tarfile.open(genomes_input[0]) as tar:
            genomes = [tarinfo.name for tarinfo in tar if tarinfo.isreg()]
    else:
        input_is_tar = False
        genomes = list(itertools.chain.from_iterable(
            ([join(g, f) for f in sorted(listdir(g))] if isdir(g) else [g])
            for g in genomes_input))
        for genome in genomes:
            if not exists(genome) or not isfile(genome):
                raise RuntimeError(f"{genome} does not exist or is not a file!")
    print_err('Recording genome sizes')
    genomes_dict = {}
    for genome in genomes:
        print_err(f'Measuring {genome}')
        genomes_dict[genome] = 0
        if input_is_tar:
            tar = tarfile.open(genomes_input[0])
            infile = gzip_agnostic_open(genome, 'rt', tar=tar)
        else:
            infile = gzip_agnostic_open(genome, 'rt')
        for record in SeqIO.parse(infile, 'fasta'):
            genomes_dict[genome] += len(str(record.seq))
        infile.close()
        if input_is_tar:
            tar.close()
    mem_blocks = generate_mem_blocks(split_memory, threads)
    all_core_blocks = []
    for m in range(1, len(mem_blocks)):
        temp_core_block = [mem_blocks[m-1][-1]]
        for i in range(len(mem_blocks[m])):
            temp_core_block.append(mem_blocks[m][i])
            all_core_blocks.append(temp_core_block)
            temp_core_block = [mem_blocks[m][i]]
    return {'kmer_size': kmer_size(),
        'version': __version__,
        'genomes': {c:g for c, g in enumerate(genomes)},
        'genome_sizes': genomes_dict,
        'positions': {},
        'mem_blocks': all_core_blocks
    }



#===============================================================================
# collect.py
#===============================================================================

"""Calculate and plot collection curves"""

# Imports ======================================================================

import os.path
import pandas as pd
import seaborn as sns
from collections import Counter
import numpy as np
from math import prod, floor
from itertools import cycle
from scipy.stats import hypergeom



# Constants ====================================================================

COL_COLOR_PALETTE = 'mako_r'




# Functions ====================================================================

def col_values_conf(pk_results):
    pkr = pk_results
    sc_matrix = np.zeros((pkr.number_of_genomes,pkr.number_of_genomes))
    for _, _, score in pkr:
        sc_matrix[sum(score) - 1,:] += np.array(score[::-1])
    sc_df = pd.DataFrame(sc_matrix, index=range(1, pkr.number_of_genomes+1),
        columns=tuple(str(os.path.basename(g)).replace('.fasta.gz', '').replace('.fa.gz', '')
                      for g in reversed(pkr.genomes)))
    g = len(sc_df.columns)
    coef_matrix_total = np.array([[prod((g-s-i)/(g-1-i) for i in range(n)) for n in range(g)]
                            for s in range(1,g+1)])
    coef_matrix_core = np.flip(coef_matrix_total.copy(), axis=0)
    for n in range(1,g):
        coef_matrix_total[:,n] += coef_matrix_total[:,n-1]
    col_df_core = pd.DataFrame(sc_df.transpose().dot(coef_matrix_core).transpose())
    col_df_total = pd.DataFrame(sc_df.transpose().dot(coef_matrix_total).transpose())
    col_df = pd.concat((col_df_total, col_df_core))
    col_df['n_genomes']= tuple(range(1,g+1))*2
    plotting_data = col_df.melt(
        id_vars=['n_genomes'], value_name='n_kmers')
    del plotting_data['variable']
    plotting_data['sequence'] = (('total',) * g + ('core',) * g) * g
    return plotting_data


def col_values(pk_results, contours=None):
    """Calculate collection curve

    Parameters
    ----------
    pk_results
        a PKResults object
    contours
        if not None, an iterable of integers between 0 and 100

    Yields
    -------
        collection curve values
    """

    pkr = pk_results
    g = pkr.number_of_genomes
    score_dist = Counter(sum(score) for _, _, score in pkr)
    for n_genomes in range(1, g+1):
        yield n_genomes, sum((1-prod((g-s-n)/(g-n) for n in range(n_genomes)))*score_dist[s]
            for s in range(1, g+1)), 'total'
        if contours:
            for c in contours:
                yield n_genomes, sum((hypergeom.sf(floor(c/100*n_genomes), g, s, n_genomes))*score_dist[s]
                    for s in range(1, g+1)), f'{c}%'
        yield n_genomes, sum(prod((s-n)/(g-n) for n in range(n_genomes))*score_dist[s]
            for s in range(1, g+1)), 'core'


def col_plot(plotting_data, output, title: str = 'Collection curve',
             linewidth: int = 3, palette=COL_COLOR_PALETTE, alpha=1, width=4,
             height=3, legend_loc='best'):
    ax = sns.lineplot(x='n_genomes', y='n_kmers', hue='sequence',
                      data=plotting_data, linewidth=linewidth,
                      palette=palette, alpha=alpha)
    ax.set_title(title)
    ax.set_ylim(bottom=0)
    if legend_loc == 'outside':
        leg = ax.legend(bbox_to_anchor=(1.02, 1), loc='upper left',
                        borderaxespad=0)
    else:
        leg = ax.legend(loc=legend_loc)
    for line in leg.get_lines():
        line.set_linewidth(linewidth)
        line.set_alpha(alpha)
    fig = ax.get_figure()
    fig.set_figwidth(width)
    fig.set_figheight(height)
    fig.tight_layout()
    fig.savefig(output)
    fig.clf()


def collect(pk_results, output=None, title: str = 'Collection curve', linewidth: int = 3,
               palette=COL_COLOR_PALETTE, alpha=1, width=4, height=3,
               conf=False, contours=False, legend_loc='best'):
    if conf:
        col_df = col_values_conf(pk_results)
    else:
        col_df = pd.DataFrame(col_values(pk_results, contours=contours),
                              columns=('n_genomes', 'n_kmers', 'sequence'))
    if output:
        col_plot(col_df, output, title=title, linewidth=linewidth,
                 palette=(palette if contours else sns.color_palette(palette, n_colors=2)), alpha=alpha,
                 width=width, height=height, legend_loc=legend_loc)
    return col_df



import pandas as pd
import upsetplot
import gzip
from collections import Counter
from itertools import compress
from matplotlib import pyplot

def generate_relevant_scores(pk_results, genomes):
    genome_set = set(genomes)
    genomes_reversed = tuple(pk_results.genomes)[::-1]
    for _, _, score in pk_results:
        memberships = set(compress(genomes_reversed, (x==1 for x in score)))
        inter = memberships.intersection(genome_set)
        if inter:
            yield tuple(sorted(inter))


def generate_relevant_scores_exclusive(pk_results, genomes):
    genome_set = set(genomes)
    genomes_reversed = tuple(pk_results.genomes)[::-1]
    for _, _, score in pk_results:
        memberships = set(compress(genomes_reversed, (x==1 for x in score)))
        inter = memberships.intersection(genome_set)
        diff = memberships - genome_set
        if inter and not diff:
            yield tuple(sorted(inter))


def count_scores(pk_results, genomes, exclusive=False):
    counts = Counter(generate_relevant_scores_exclusive(pk_results, genomes)
        if exclusive else generate_relevant_scores(pk_results, genomes))
    return pd.Series(counts.values(), index=pd.MultiIndex.from_tuples(
        (tuple(g in k for g in genomes) for k in counts.keys()),
        names=genomes))


def upset(pk_results, output, genomes, vertical=False, show_counts=False,
          min_subset_size=None, max_subset_size=None, exclusive=False,
          table=None):
    score_counts = count_scores(pk_results, genomes, exclusive=exclusive)
    if table:
        with (gzip.open if table.endswith('.gz') else open)(table, 'wb') as f:
            score_counts.to_csv(f, sep='\t', header=['k-mers'])
    upsetplot.plot(score_counts,
        orientation='vertical' if vertical else 'horizontal',
        show_counts=show_counts,
        min_subset_size=min_subset_size,
        max_subset_size=max_subset_size)
    pyplot.savefig(output)



def get_lower_bound(positions: dict, theo_bound: int) -> int:
    lower_bound = 0
    for kmer, pos in positions.items():
        if int(kmer) > theo_bound:
            return lower_bound
        lower_bound = pos



#===============================================================================
# reg_coverage.py
#===============================================================================

"""High-level functions for k-mer coverage analysis"""



# Imports ======================================================================

import gzip
import matplotlib.pyplot as plt
import os.path
import pandas as pd
import seaborn as sns
import gff2bed
from Bio.bgzf import BgzfReader, BgzfWriter
from io import IOBase
from itertools import accumulate, chain, groupby, islice, cycle
from math import floor, log10
from operator import itemgetter
from pankmer.env import GENES_PATH, COORD_REGEX
from pyfaidx import Fasta
from statistics import mean




# Constants ====================================================================

COLOR_PALETTE = sns.color_palette().as_hex()




# Functions ====================================================================

def parse_coords(coords: str):
    """Parse the coordinate string into its components
    
    Parameters
    ----------
    coords : str
        String of form chr:start-end giving genomic coordinates
    """

    chrom, start, end = coords.replace('-', ':').split(':')
    return chrom, int(start), int(end)


def parse_gene(gene: str, genes_path=GENES_PATH): # Not used yet, but will be useful later
    with gzip.open(genes_path, 'rt') as f:
        for line in f:
            parsed_line = line.split()
            if parsed_line[3] == gene:
                chrom, start, end = parsed_line[:3]
                break
        else:
            raise RuntimeError('gene not found')
    return chrom, int(start), int(end)


def gff_to_dict(gff, type='gene', n_features=None):
    """Parse a GFF file into a dictionary of feature coordinates

    Parameters
    ----------
    gff: str
       path to GFF file
    type
        string indicating feature type to include, or None to include all
        features
    n_features
        int indicating number of features to include, or None to include all
        features

    Returns
    -------
    dict
        coordinates of features in the GFF
    """

    sorted_features = sorted(gff2bed.parse(gff, type=type, parse_attr=False),
                             key=itemgetter(1))
    if n_features:
        return {chrom: [[start, end] for _, start, end, *_ in coords][:n_features]
            for chrom, coords in groupby(sorted_features, key=itemgetter(0))}
    else:
        return {chrom: [[start, end] for _, start, end, *_ in coords]
            for chrom, coords in groupby(sorted_features, key=itemgetter(0))}


def attr_to_coords(gff, type='gene', attr="Name"):
    """Parse a GFF file into a dictionary of feature coordinates with an attr
       (e.g. "Name") as keys

    Parameters
    ----------
    gff: str
       path to GFF file
    type
        string indicating feature type to include, or None to include all
        features
    n_features
        int indicating number of features to include, or None to include all
        features

    Returns
    -------
    dict
        coordinates of features in the GFF keyed by attr 
    """

    parsed_features = gff2bed.parse(gff, type=type)
    return {attr_vals[attr]: (seqid, start, end)
        for seqid, start, end, _, attr_vals in parsed_features}


def generate_coverage_values(chrom, start, end, *regcov_dicts, summary_func=mean):
    for position, *values in zip(range(start, end + 1),
        *(next(iter(regcov_dict[chrom].values())) for regcov_dict in regcov_dicts)):
        yield chrom, position-1, position, *(summary_func(v) for v in values)


def reg_coverage(*pk_results, ref, coords, summary_func=mean, output_file=None,
                 bgzip: bool = False, genes=None, flank: int = 0,
                 processes: int = 1):
    """Generate k-mer coverage levels across the input region

    Parameters
    ----------
    pk_results
        a PKResults object
    ref: str
        path to a reference genome in BGZIP compressed FASTA format
    coords: str
        genomic coordinates formatted as ctg:start-end
    summary_func
        function for summarizing the coverage value of a kmer across genomes
        in the index. The default is statistics.mean
    output_file
        filename or file object to write to
    bgzip: bool
        if True, output_file will be block compressed
    genes: str
        path to gff3 file containing gnes
    flank: int
        size of flanking region
    processes: int
        number of processes to use

    Yields
    ------
    contig, start, end, *values
        row of bedGraph data
    """
    
    if genes and not (COORD_REGEX.match(coords)):
        contig, start, end = attr_to_coords(genes)[coords]
    else:
        contig, start, end = parse_coords(coords)
    start -= flank
    end += flank
    for result in pk_results:
        result.threads = processes
    regcov_dicts = (result.get_collapsed_regional_scores(ref, {contig: [[start, end]]})
                    for result in  pk_results)
    if output_file:
        with (output_file if isinstance(output_file, (BgzfWriter, IOBase))
              else (BgzfWriter if bgzip else open)(output_file, 'wb')) as f:
            for cg, st, ed, *vs in generate_coverage_values(
                contig, start, end, *regcov_dicts, summary_func=summary_func
            ):
                f.write(('\t'.join(str(x) for x in (cg, st, ed, *vs))+'\n').encode())
    return generate_coverage_values(contig, start, end, *regcov_dicts,
                                        summary_func=summary_func)


def check_for_bgzip(reference: str):
    try:
        BgzfReader(reference)
    except ValueError:
        raise RuntimeError('Input FASTA must be BGZIP compressed')        


def get_chromosome_sizes_from_ref(reference: str):
    """Extract chromosome sizes from reference FASTA

    Parameters
    ----------
    reference : str
        path to reference FASTA file

    Returns
    -------
    DataFrame
        name and size of each chromosome
    """
    return pd.DataFrame(((k, len(v)) for k, v in Fasta(reference).items()),
                        columns=('name', 'size'))



def generate_plotting_data(coverage_values, groups, size, scale: float = 1,
                           shift: float = 0, bin_size: int = 0,
                           single_chrom: bool = False):
    """Construct rows of preprocessed data for the plotting data frame. Data
    are binned by rounding to the nearest bin coordinate, while bin coordinates
    are determined by the bin size parameter.

    Parameters
    ----------
    coverage values
        iterable of iterables listing coverage values
    groups
        iterable of group names
    size
        chromosome size in bp
    scale
        ratio of chromosome size to mean chromosome size
    shift
        x-axis shift of this chromosome, for plots showing multiple chromosomes
        consecutively
    bin_size
        set bin size. The input <int> is converted to the bin size by the
        formula: 10^(<int>+6) bp. The default value is 0, i.e. 1-megabase bins.

    Yields
    ------
    tuple
        bin coordinate, value, and group ID of a coverage data point
    """

    yield from ((chrom, min(round(int(pos), -6-bin_size), size)/size*scale + shift,
                float(cov), group, f'{group}' if single_chrom else f'{group}_{chrom}')
        for chrom, _, pos, cov, group
        in (list(vals)+[g] for cv, g in zip(coverage_values, groups) for vals in cv))


def collapse_plotting_data(rows):
    yield from ((c, x, mean(r[2] for r in rws), g, n)
                for (c, x, g, n), rws in groupby(rows, key=itemgetter(0,1,3,4)))


def regcov_dicts_tuple(*pk_results, ref, chromosomes):
    return tuple(result.get_collapsed_regional_scores(ref, {c: [] for c in chromosomes})
                    for result in  pk_results)


def genome_coverage_df(*regcov_dicts, sizes, chromosomes: list,
                       summary_func=mean, groups=None,
                       legend_title: str = 'Group', bin_size: int = 0,
                       x_label: str = 'Chromosome'):
    sizes.index = sizes.name
    sizes = sizes.loc[chromosomes, 'size']
    scales = sizes / sizes.mean()
    shifts = pd.Series(accumulate(chain((0,),scales[:-1])), index=scales.index)
    return pd.DataFrame(
        chain.from_iterable((collapse_plotting_data(generate_plotting_data(
                (generate_coverage_values(chrom, 0,
                    sizes[chrom], rd, summary_func=summary_func)
                    for rd in regcov_dicts),
                groups, size, scale=scale, shift=shift, bin_size=bin_size,
                single_chrom=(len(chromosomes)==1)))
            for chrom, size, scale, shift
            in zip(chromosomes, sizes, scales, shifts))),
        columns=('SeqID', x_label, 'K-mer coverage (%)', legend_title,
                 f'{legend_title}_chrom'))


def genome_coverage_plot(plotting_data, output: str,
                         groups=None, loci=None, sizes=None,
                         title: str = 'Coverage', x_label: str = 'Chromosome',
                         legend: bool = False,
                         legend_title: str = 'Group', legend_loc: str = 'best',
                         width: float = 7.0, height: float = 3.0,
                         color_palette=COLOR_PALETTE, alpha: float = 0.5,
                         linewidth: int = 3, xtlabel_rotation: int = 0,
                         xtlabel_ha: str = 'center'):
    """Generate a plot of average k-mer coverage values in bins, from a DF
    generated by genome_coverage_df

    Parameters
    ----------
    plotting_data
        pandas DataFrame as generated by genome_coverage_df
    output : str
        path to destination file for the plot
    groups
        iterable of group names
    loci
        list of strings indicating loci in form "contig:pos:name"
    sizes
        pandas DataFrame containing chrom.sizes table
    title : str
        title of the plot
    x_label : str
        x-axis label
    legend : bool
        if true, draw a legend for the plot
    legend_title : str
        title for the plot legend
    legend_loc : str
        location of legend. must be one of "best", "upper left", "upper right",
        "lower left", "lower right", "outside"
    width : float
        width of the plot in inches
    height : float
        height of the plot in inches
    color_palette
        color palette for plot lines
    alpha : float
        alpha value for plot lines
    linewidth : float
        width value for plot lines
    xtlabel_rotation : int
        rotation for x-axis tick labels
    xtlabel_ha : str
        horizontal alignment for x-axis tick labels (left, center, right)
    """

    chromosomes = plotting_data.loc[:,'SeqID'].unique()
    if legend:
        data_groups = plotting_data.loc[:,legend_title].unique()
        if groups is None:
            groups = data_groups
        else:
            groups_dict = dict(zip(data_groups, groups))
            plotting_data[legend_title] = tuple(groups_dict[x] for x in plotting_data[legend_title])
            plotting_data[f'{legend_title}_chrom'] = tuple(groups_dict[x] for x in plotting_data[f'{legend_title}_chrom'])
    palette = tuple(islice(cycle(color_palette[:len(groups)]),
                                 len(chromosomes) * len(groups)))
    shifts = plotting_data.loc[:,['SeqID', x_label]].groupby('SeqID').min()
    if sizes is not None:
        sizes.index = sizes.name
        sizes = sizes.loc[chromosomes, 'size']
    if loci is not None:
        if sizes is None:
            raise RuntimeError('loci argument requires sizes argument')
        scales = sizes / sizes.mean()
        loci_parsed = tuple(
            (int(p)/sizes[c] * scales[c] + shifts.loc[c, x_label], n)
            for c, p, n in (l.split(':') for l in loci))
        loci_x = tuple(l[0] for l in loci_parsed)
        ticks_labels = sorted(loci_parsed + tuple(zip(shifts[x_label],
                                                      chromosomes)))
        xticks = tuple(tl[0] for tl in ticks_labels)
        xlabels = tuple(tl[1] for tl in ticks_labels)
    else:
        xticks = shifts[x_label]
        xlabels = chromosomes
    ax = sns.lineplot(x=x_label, y='K-mer coverage (%)',
                      hue=f'{legend_title}_chrom', data=plotting_data, errorbar=None,
                      linewidth=linewidth,
                      palette=palette,
                      alpha=alpha,
                      legend='auto' if legend else False)
    ax.set_title(title)
    if (loci is not None) and (sizes is not None):
        ax.vlines(x=loci_x, ymin=min(plotting_data['K-mer coverage (%)']),
                  ymax=max(plotting_data['K-mer coverage (%)']),
                  colors='gray',
                  linestyles='dashed')
    if (len(chromosomes) == 1) and (sizes is not None):
        xticks = ax.get_xticks()[1:-1]
        xlabels=tuple(f"{x*sizes.loc[chromosomes[0]]/1e6:.1f}" for x in xticks)
    ax.set_xticks(xticks)
    ax.set_xticklabels(xlabels, rotation=xtlabel_rotation, ha=xtlabel_ha)
    if legend:
        if legend_loc == 'outside':
            leg = ax.legend(bbox_to_anchor=(1.02, 1), loc='upper left',
                            borderaxespad=0, title=legend_title)
        else:
            leg = ax.legend(loc=legend_loc, title=legend_title)
        for line in leg.get_lines():
            line.set_linewidth(linewidth)
            line.set_alpha(alpha)
    fig = ax.get_figure()
    fig.set_figwidth(width)
    fig.set_figheight(height)
    fig.tight_layout()
    fig.savefig(output)
    fig.clf()


def genome_coverage(*pk_results, output: str, ref: str, chromosomes: list,
                    output_table=None, summary_func=mean, groups=None, loci=None,
                    title: str = 'Coverage', x_label: str = 'Chromosome',
                    legend: bool = False, legend_title='Group', legend_loc='best',
                    bin_size: int = 0, width: float = 7.0, height: float = 3.0,
                    color_palette=COLOR_PALETTE, alpha: float = 0.5,
                    linewidth: int = 3, processes: int = 1,
                    xtlabel_rotation: int = 0, xtlabel_ha: str = 'center'):

    """Generate a plot of average k-mer coverage values in bins across the
    input genome

    Parameters
    ----------
    pk_results
        a PKResults object
    output : str
        path to destination file for the plot
    ref : str
        path to the input reference genome in BGZIP compressed FASTA format
    chromosomes : list
        list of chromosomes to include in the plot
    output_table
        path to write TSV table containing underlying plotting data
    summary_func
        function for summarizing the coverage value of a kmer across genomes
        in the index. The default is statistics.mean
    groups
        iterable of group names
    title : str
        title of the plot
    x_label : str
        x-axis label
    legend : bool
        if true, draw a legend for the plot
    legend_title : str
        title for the plot legend
    legend_loc : str
        location of legend. must be one of "best", "upper left", "upper right",
        "lower left", "lower right", "outside"
    bin_size
        set bin size. The input <int> is converted to the bin size by the
        formula: 10^(<int>+6) bp. The default value is 0, i.e. 1-megabase bins.
    width : float
        width of the plot in inches
    height : float
        height of the plot in inches
    color_palette
        color palette for plot lines
    alpha : float
        alpha value for plot lines
    linewidth : float
        width value for plot lines
    processes: int
        number of processes to use
    xtlabel_rotation : int
        rotation for x-axis tick labels
    xtlabel_ha : str
        horizontal alignment for x-axis tick labels (left, center, right)
    """

    check_for_bgzip(ref)
    if not groups:
        groups = list(range(len(pk_results)))
    for result in pk_results:
        result.threads = processes
    regcov_dicts = regcov_dicts_tuple(*pk_results, ref=ref,
        chromosomes=chromosomes)
    sizes = get_chromosome_sizes_from_ref(ref)
    plotting_data = genome_coverage_df(*regcov_dicts, sizes=sizes,
        chromosomes=chromosomes, summary_func=summary_func, groups=groups,
        legend_title=legend_title, bin_size=bin_size, x_label=x_label)
    if output_table:
        plotting_data.to_csv(output_table, sep='\t', index=False)
    genome_coverage_plot(plotting_data, output=output,
        groups=groups, loci=loci, sizes=sizes, title=title, x_label=x_label,
        legend=legend, legend_title=legend_title, legend_loc=legend_loc,
        width=width, height=height, color_palette=color_palette, alpha=alpha,
        linewidth=linewidth, xtlabel_rotation=xtlabel_rotation,
        xtlabel_ha=xtlabel_ha)


def coverage_heatmap(pk_results, refs, features, output, n_features=None, width=7, height=7):
    if not output.endswith('.png'):
        raise RuntimeError('output file path must end with .png')
    results = tuple(pk_results.get_collapsed_regional_scores(ref, gff_to_dict(gff, n_features=n_features))
        for ref, gff in zip(refs, features))
    dfs = tuple(pd.DataFrame({os.path.basename(ref.replace('.fasta', '').replace('.gz', '')):
            tuple(sum(v) for values in regcov_dict.values() for v in values)}).transpose()
        for ref, regcov_dict in zip(refs, results))
    vmin = min(df.values.min() for df in dfs)
    vmax = max(df.values.max() for df in dfs)
    sns.set_context('paper')
    fig, axs = plt.subplots(nrows=len(dfs))
    fig.set_figwidth(width)
    fig.set_figheight(height)
    def round(n, k):
        return n - n % k
    for df, ax in zip(dfs, axs):
        dft = df.transpose()
        tick_step = int(10**(floor(log10(dft.index.max() - dft.index.min()))))
        tick_min = int(round(dft.index.min(), (-1 * tick_step)))
        tick_max = int(round(dft.index.max(), (1 * tick_step))) + tick_step
        xticklabels = range(tick_min, tick_max, tick_step)
        xticks = [dft.index.get_loc(label) for label in xticklabels]
        xticklabels_pretty = tuple(
            (f'{int(x)}BP' if x<1e3
             else f'{int(x/1e3)}KB' if x<1e6
             else f'{int(x/1e6)}MB')
            for x in xticklabels)
        sns.heatmap(df, ax=ax, vmin=vmin, vmax=vmax,
            xticklabels=xticklabels_pretty)
        ax.tick_params(labelrotation=0)
        ax.set_xticks(xticks, rotation=0, labels=xticklabels_pretty)
    fig.tight_layout()
    fig.savefig(output)



import numpy as np
import pandas as pd
import sys
from itertools import chain
from scipy.cluster.hierarchy import linkage
import newick as nw

def get_kmers_per_col(mat):
    """From an adjacency matrix, generate a square matrix where each element
    of each column is the total number of K-mers in the genome indexed at that
    column.

    This is a convenience function used to generate a data structure that
    makes it easier to do vectorized computation of jaccard and overlap
    similarities.
    
    Parameters
    ----------
    mat
        a numpy array containing he adjacency matrix
    
    Returns
    -------
    array
        a numpy array containing the "K-mers per column" matrix
    """

    kmers_per_genome = np.diagonal(mat)
    return np.array([kmers_per_genome for _ in range(kmers_per_genome.shape[0])])


def adj_to_jaccard(adj_matrix):
    """Convert a raw adjacency matrix to Jaccard similarity values
    
    Parameters
    ----------
    adj_matrix
        pandas data frame containing the adjacency matrix
    
    Returns
    -------
    DataFrame
        pandas data framed containing the jaccard matrix
    """

    mat = adj_matrix.to_numpy()
    kmers_per_col = get_kmers_per_col(mat)
    jaccard = mat / (kmers_per_col + np.transpose(kmers_per_col) - mat)
    return pd.DataFrame(jaccard, index=adj_matrix.index, columns=adj_matrix.columns)


def adj_to_overlap(adj_matrix):
    """Convert a raw adjacency matrix to overlap similarity values
    
    Parameters
    ----------
    adj_matrix
        pandas data frame containing the adjacency matrix
    
    Returns
    -------
    DataFrame
        pandas data framed containing the overlap matrix
    """

    mat = adj_matrix.to_numpy()
    kmers_per_col = get_kmers_per_col(mat)
    overlap = mat / np.minimum(kmers_per_col, np.transpose(kmers_per_col))
    return pd.DataFrame(overlap, index=adj_matrix.index, columns=adj_matrix.columns)


def compute_linkage_matrix(adj_matrix, method='complete', optimal_ordering=True):
    """Compute a linkage matrix defining a hierarchical clustering
    
    Parameters
    ---------
    adj_matrix
        pandas data frame containing the adjacency matrix
    method : str
        linkage algorithm for hierarchical clustering. See
        `scipy.cluster.hierarchy.linkage` for details.
    optimal_ordering : bool
        If True, the linkage matrix will be reordered so that the distance
        between successive leaves is minimal. See
        `scipy.cluster.hierarchy.linkage`
    """

    adj_matrix_np = adj_matrix.to_numpy()
    similarity = adj_matrix_np[np.triu_indices(len(adj_matrix.index), k=1)]
    distance = adj_matrix_np.max() - similarity
    return linkage(distance, method=method, optimal_ordering=optimal_ordering)


def linkage_df(linkage_matrix):
    """Convert a linkage matrix to linkage data frame.
    This allows different dtypes and column headers.

    Parameters
    ----------
    linkage_matrix
        linkage matrix generated by `scipy.cluster.hierarchy.linkage`
    
    Returns
    -------
    DataFrame
        data frame containing the linkage matrix
    """

    df = pd.DataFrame(linkage_matrix)
    for n in 0, 1, 3:
        df[n] = df[n].astype(int)
    df.columns = 'clust_0', 'clust_1', 'dist', 'n_obs'
    return df


def linkage_to_newick(linkage_df, names):
    """Generate a NEWICK formatted tree from a linkage data frame

    Parameters
    ----------
    linkage_df
        pandas data frame containing the linkage matrix
    names
        iterable of genome names corresponding to clusters in linkage_df
    
    Returns
    -------
    str
        NEWICK formatted tree
    """
    
    n_genomes = len(names)
    names_dict = dict(enumerate(names))
    tree = list(chain.from_iterable(
        (nw.Node(name=names_dict.get(clust_0, str(clust_0)), length=str(dist)),
         nw.Node(name=names_dict.get(clust_1, str(clust_1)), length=str(dist)))
        for _, clust_0, clust_1, dist, _ in linkage_df.itertuples()))
    tree.append(nw.Node(name=str(2*n_genomes-2), length=tree[-1]._length))
    nodes_dict = {n.name: n for n in tree}
    for i in range(n_genomes - 1):
        nodes_dict[str(i+n_genomes)].add_descendant(tree[2*i])
        nodes_dict[str(i+n_genomes)].add_descendant(tree[2*i+1])
        l0 = float(nodes_dict[str(i+n_genomes)]._length)
        l1 = float(tree[2*i]._length)
        nodes_dict[str(i+n_genomes)]._length = str(l0 - l1)
    tree[-1].name = None
    tree[-1]._length = None
    return nw.dumps(tree[-1])


def tree(adj_matrix, newick: bool = False, metric: str = 'intersection',
         method: str = 'complete', optimal_ordering: bool = True,
         transformed_matrix = None):
    """Generate a hierarchical clustering tree from an adjacency matrix

    Parameters
    ----------
    adj_matrix
        a pandas data frame containing an adjacency matrix
    newick : bool
        if True, print a NEWICK-formatted tree. Otherwise, print a linkage
        matrix
    metric : str
        similarity metric to use for clustering. must be "intersection",
        "jaccard", or "overlap"
    method : str
        linkage algorithm for hierarchical clustering. See
        `scipy.cluster.hierarchy.linkage` for details.
    optimal_ordering : bool
        If True, the linkage matrix will be reordered so that the distance
        between successive leaves is minimal. See
        `scipy.cluster.hierarchy.linkage`
    """
    
    if metric == 'jaccard':
        adj_matrix = adj_to_jaccard(adj_matrix)
    elif metric == 'overlap':
        adj_matrix = adj_to_overlap(adj_matrix)
    if transformed_matrix:
        with open(transformed_matrix, 'w') as f:
            adj_matrix.to_csv(f)
    linkage_matrix = compute_linkage_matrix(adj_matrix, method=method,
                                     optimal_ordering=optimal_ordering)
    df = linkage_df(linkage_matrix)
    if newick:
        newick_tree = linkage_to_newick(df, adj_matrix.index)
        print(newick_tree)
    else:
        df.to_csv(sys.stdout, index=False, sep='\t')



import numpy as np
import os.path
import pandas as pd
from pankmer.index import add_score_mat_np

# def get_adjacency_matrix(input_dir):
#     metadata_file = join(input_dir, 'metadata.json')
#     with open(metadata_file, 'rt') as infile:
#         metadata = json.load(infile)
#     k = metadata['kmer_size']
#     genomes = list(metadata['genomes'].keys())
#     genomes_count = len(genomes)
#     score_bitsize = math.ceil(genomes_count/8)
#     scores_counts = count_scores(join(input_dir, 'scores.b.gz'), score_bitsize)
#     dim = len(genomes)
#     mat = np.zeros((dim, dim), dtype=int)
#     for count, score in enumerate(scores_counts):
#         score_int = int.from_bytes(score, 'big', signed=False)
#         score_multi = np.array(list(bin(int(score_int))[2:]), dtype=int)*scores_counts[score]
#         mat = add_score_mat_np(score_multi, mat)
#     df = pd.DataFrame(mat, index=genomes[::-1], columns=genomes[::-1])
#     return df

def get_adjacency_matrix(results):
    """Generate an adjacency matrix from a PKResults object
    
    Parameters
    ----------
    results
        a PKResults object
    
    Returns
    -------
    DataFrame
        an adjacency matrix    
    """
    pass
    score_counts = {}
    for _, _, score in results:
        score_counts[score] = score_counts.get(score, 0) + 1
    print(score_counts)
    mat = np.zeros((results.number_of_genomes,
                    results.number_of_genomes), dtype=int)
    for score, count in score_counts.items():
        score_multi = np.array(score) * count
        mat = add_score_mat_np(score_multi, mat)
    genome_names = tuple(str(os.path.basename(g)).replace('.fasta.gz', '').replace('.fa.gz', '')
                         for g in results.genomes)
    return pd.DataFrame(mat, index=genome_names, columns=genome_names)



from pankmer.version import __version__
from pankmer.pankmer import PKResults
from pankmer.collect import collect
from pankmer.adjacency_matrix import get_adjacency_matrix
from pankmer.clustermap import clustermap
from pankmer.reg_coverage import genome_coverage, reg_coverage



import pandas as pd
from collections import Counter

def count_kmers(*pk_results, names=None):
    """Count total and diagnostic K-mers in one or more indexes

    Parameters
    ----------
    pk_results
        a PKResults object
    names
        an iterable of the names for each input index

    Returns
    -------
    DataFrame
        table of K-mer counts for each index
    """
    names = names or range(len(pk_results))
    max_score = max(pkr.number_of_genomes for pkr in pk_results)
    kmer_counts = pd.DataFrame(0, columns=names, index=range(1, max_score+1))
    for pkr, name in zip(pk_results, names):
        for s, c in  Counter(sum(score) for _, _, score in pkr).items():
            kmer_counts.loc[s, name] = c
    return kmer_counts



import gzip
import os
import os.path
import io
import math
import tarfile
import shutil
import json
from itertools import compress
from pankmer.index import load_scores, dump_scores, subset_scores
from pankmer.version import __version__

def bools_to_score(bools):
    return int(''.join(str(int(b)) for b in bools), base = 2)


def generate_subset_scores(pk_results, genomes):
    genome_set = set(genomes)
    for kmer, i, score in pk_results:
        memberships = set(compress(pk_results.genomes, (x==1 for x in score)))
        inter = memberships.intersection(genome_set)
        if inter:
            yield kmer, i
            # yield kmer, i, bools_to_score(g in memberships for g in genomes)


def generate_subset_scores_exclusive(pk_results, genomes):
    genome_set = set(genomes)
    for kmer, i, score in pk_results:
        memberships = set(compress(pk_results.genomes, (x==1 for x in score)))
        inter = memberships.intersection(genome_set)
        diff = memberships - genome_set
        if inter and not diff:
            yield kmer, i
            # yield kmer, i, bools_to_score(g in memberships for g in genomes)


def subset(pk_results, output, genomes, exclusive=False):
    metadata_dict = {
        "kmer_size": pk_results.kmer_size,
        "version": __version__,
        "genomes": dict(enumerate(genomes)),
        "genome_sizes": {g: pk_results.genomes[g] for g in genomes},
        "positions": {},
        "mem_blocks": pk_results.mem_blocks
    }
    kmer_bitsize = math.ceil(((pk_results.kmer_size*2))/8)
    i_bitsize = 8
    output_is_tar = output.endswith('.tar')
    output_dir = output[:-4] if output_is_tar else output
    if not os.path.exists(output_dir):
        os.mkdir(output_dir)
    idx_map = subset_scores(pk_results.results_dir,
        (pk_results.results_dir if pk_results.input_is_tar else ''),
        tuple(pk_results.genomes), genomes, os.path.join(output_dir, 'scores.pks'))
    idx_map_bytes = {i.to_bytes(i_bitsize, byteorder="big", signed=False):
                         j.to_bytes(i_bitsize, byteorder="big", signed=False)
                     for i, j in idx_map.items()}
    kmers_out_path = os.path.join(output_dir, f'kmers.bgz')
    indices_out_path = os.path.join(output_dir, f'indices.bgz')
    with gzip.open(kmers_out_path, 'wb') as kmers_out, gzip.open(indices_out_path, 'wb') as indices_out:
        with io.BufferedWriter(indices_out, buffer_size=1000*i_bitsize) as io_buffer ,\
            io.BufferedWriter(kmers_out, buffer_size=1000*kmer_bitsize) as ko_buffer:
            count = 0
            for kmer, i in (generate_subset_scores_exclusive(pk_results, genomes) if exclusive
                else generate_subset_scores(pk_results, genomes)):
                ko_buffer.write(kmer)
                io_buffer.write(idx_map_bytes[i])
                if count%10000000 == 0 and count != 0: 
                    metadata_dict['positions'][str(pk_results.decode_kmer(kmer))] = count
                    count = 0
                count += 1
    with open(f'{output_dir}/metadata.json', 'w') as f:
        json.dump(metadata_dict, f)
    if output_is_tar:
        with tarfile.open(output, 'w') as tar:
            tar.add(output_dir)
        shutil.rmtree(output_dir)



import pandas as pd
import seaborn as sns
from pankmer.tree import adj_to_jaccard, adj_to_overlap, compute_linkage_matrix

def clustermap(adj_matrix, output: str, cmap: str = 'mako_r',
               width: float = 7.0, height: float = 7.0,
               metric='intersection', method='complete',
               optimal_ordering: bool = True, square: bool = True,
               heatmap_tick_pos='right', cbar_tick_pos='left',
               dendrogram_ratio: float = 0.2, dendrogram_spacer:float = 0.1):
    """Plot a clustered heatmap of genome similarity values

    Parameters
    ----------
    adj_matrix : DataFrame
        a Pandas data frame representing the adjacency matrix
    output : str
        path to destination file for plot
    cmap : str
        color map for the heatmap
    width : float
        width of the plot in inches
    height : float
        height of the plot in inches
    metric : str
        similarity metric to use, chosen from "intersection", "jaccard", or
        "overlap"
    method : str
        linkage algorithm for hierarchical clustering. See
        `scipy.cluster.hierarchy.linkage` for details.
    optimal_ordering : bool
        If True, the linkage matrix will be reordered so that the distance
        between successive leaves is minimal. See
        `scipy.cluster.hierarchy.linkage`
    square : bool
        If True, the heatmap will be drawn square. This may force the figure
        dimensions to be altered slightly
    cbar_tick_pos : str
        Position of color bar ticks. Must be "left" or "right" [left]
    dendrogram_ratio : float
        Fraction of plot width used for dendrogram [0.2]
    dendrogram_spacer : float
        Fraction of plot width used as spacer between dendrogram and heatmap [0.1]
    """
    
    if metric == 'jaccard':
        adj_matrix = adj_to_jaccard(adj_matrix)
    elif metric == 'overlap':
        adj_matrix = adj_to_overlap(adj_matrix)
    linkage = compute_linkage_matrix(adj_matrix, method=method,
                                     optimal_ordering=optimal_ordering)
    ax = sns.clustermap(adj_matrix, cmap=cmap, figsize=(width, height),
                        row_linkage=linkage, col_linkage=linkage,
                        xticklabels=False, dendrogram_ratio=(dendrogram_ratio,0))
    if square:
        ax.ax_heatmap.set_aspect("equal")
    ax.ax_col_dendrogram.set_visible(False)
    ax_hm_pos = ax.ax_heatmap.get_position()
    ax_row_pos = ax.ax_row_dendrogram.get_position()
    match heatmap_tick_pos:
        case 'left':
            ax.ax_heatmap.set_position([.9-ax_hm_pos.width, ax_hm_pos.y0,
                                        ax_hm_pos.width, ax_hm_pos.height])
        case 'right':
            ax.ax_heatmap.set_position([ax_hm_pos.x0-.1, ax_hm_pos.y0,
                                        ax_hm_pos.width, ax_hm_pos.height])
    ax.ax_heatmap.yaxis.set_ticks_position(heatmap_tick_pos)
    ax.ax_heatmap.set_yticklabels(ax.ax_heatmap.get_yticklabels(), rotation=0)
    ax.ax_row_dendrogram.set_position([ax_row_pos.x0, ax_hm_pos.y0, 
                                       ax_row_pos.width-dendrogram_spacer, ax_hm_pos.height])
    ax.ax_cbar.set_position([.97, ax_hm_pos.y0, .03, ax_hm_pos.height])
    ax.ax_cbar.yaxis.set_ticks_position(cbar_tick_pos)
    ax.savefig(output)



import gzip
from io import BytesIO, TextIOWrapper

def nothing(x):
    return x

def gzip_agnostic_open(file_path, mode='rb', tar=None, **kwargs):
    return (gzip.open(tar.extractfile(file_path) if tar is not None
                      else file_path, mode, **kwargs)
            if str(file_path).endswith('.gz')
            else ((lambda b: TextIOWrapper(b) if mode[1]=='t' else b)(
                   BytesIO(tar.extractfile(file_path).read()), **kwargs)
                  if tar is not None else open(file_path, mode, **kwargs)))


#!/usr/bin/env python
import argparse
import gzip
import io
from os.path import join, isfile, isdir, basename, dirname, exists
import seaborn as sns
from os import mkdir, listdir
import sys
import time
from multiprocessing import Pool
import pandas as pd
import math
import json
from Bio import SeqIO
import numpy as np
import shutil
import tarfile
import itertools
from operator import itemgetter

# from pankmer.index import (print_err, genome_name, break_seq, get_kmers,
#                    score_byte_to_blist, kmer_byte_to_blist, kmer_byte_to_long)
from pankmer.index import (kmer_size, break_seq, print_err, score_byte_to_blist,
                           run_index, load_scores)
from pankmer.env import EXAMPLE_DATA_DIR
from pankmer.download import download_example
from pankmer.count import count_kmers
from pankmer.collect import COL_COLOR_PALETTE, collect
from pankmer.upset import upset
from pankmer.subset import subset
from pankmer.adjacency_matrix import get_adjacency_matrix
from pankmer.tree import tree
from pankmer.clustermap import clustermap
from pankmer.reg_coverage import (COLOR_PALETTE, reg_coverage, genome_coverage,
                                  genome_coverage_plot, coverage_heatmap)
from pankmer.dryrun import dryrun
from pankmer.version import __version__
from pankmer.gzip_agnostic_open import gzip_agnostic_open
from pankmer.get_lower_bound import get_lower_bound

#===============================================================================


def retrieve_metadata(metadata_path):
    if isinstance(metadata_path, tarfile.ExFileObject):
        metadata = json.load(metadata_path)
    else:
        with open(metadata_path, 'rt') as infile:
            metadata = json.load(infile)
    genomes = {}
    for pos in range(len(metadata['genomes'])):
        genome = metadata['genomes'][str(pos)]
        size = metadata['genome_sizes'][genome]
        genomes[genome] = size
    return metadata['kmer_size'], genomes, metadata['positions'], metadata['mem_blocks']


def get_positional_kmers(fasta_path, upper, lower, contig_header='id'):
    kmers_dict = {}
    with gzip_agnostic_open(fasta_path, 'rt') as infile:
        for record in SeqIO.parse(infile, 'fasta'):
            names = {'id': record.id, 'description': record.description}
            seq_byte = bytes(str(record.seq).upper(), 'ascii')
            kmers_dict[names[contig_header]] = break_seq(seq_byte, upper, lower)
    return kmers_dict


def get_sorted_kmer_scores(args: list) -> dict:
    index_path, kmers = args
    results = PKResults(index_path)
    lower = min(list(kmers.keys()))
    upper = max(list(kmers.keys()))
    lb_bit = int(lower).to_bytes(results.kmer_bitsize, byteorder="big", signed=False) # lower bound bit
    ub_bit = int(upper).to_bytes(results.kmer_bitsize, byteorder="big", signed=False)
    # Find lower bound position from index
    initial_lower_bound = get_lower_bound(results.positions, lower)
    # Move iterator from initial lower bound to actual lower bound
    if initial_lower_bound != None:
        results.seek_kmer(initial_lower_bound)
        results.seek_i(initial_lower_bound)
    for kmer_bit, _, score in results:
        if kmer_bit > ub_bit:
            break
        kmer  = results.decode_kmer(kmer_bit)
        if kmer in kmers:
            kmers[kmer] = score
    return kmers


class PKIterator:
    def __init__(self, results):
        self._results = results
    
    def __next__(self):
        kmer = self._results.get_kmer_byte()
        i_score = self._results.get_score_byte()
        if kmer == None or i_score == None:
            self._results.reset_iter()
            raise StopIteration
        return (kmer, *i_score)


class PKResults:
    def __init__(self, results_dir: str, threads: int = 1):
        index_files = ['metadata.json', 'kmers.bgz', 'indices.bgz', 'scores.pks']
        self.input_is_tar = False
        if isfile(results_dir) and tarfile.is_tarfile(results_dir):
            self.input_is_tar = True
            self.tar = tarfile.open(results_dir)
            found_files = {basename(tarinfo.name) for tarinfo in self.tar
                            if tarinfo.isreg()}
            for file in index_files:
                if file not in found_files:
                    raise ValueError(f"{file} is not found in index!")
            indices_filename, kmers_filename, metadata_filename, scores_filename = sorted(
                tarinfo.name for tarinfo in self.tar if tarinfo.isreg())
            kmer_size, genomes, positions, mem_blocks = retrieve_metadata(
                self.tar.extractfile(metadata_filename))
            self.number_of_genomes = len(genomes)
            self.scores = {i: tuple(self.decode_score(bytes(s))) for i, s in enumerate(load_scores(str(results_dir), str(results_dir)))}
            self.kmers_stream = gzip.open(self.tar.extractfile(kmers_filename), 'rb')
            self.indices_stream = gzip.open(self.tar.extractfile(indices_filename), 'rb')
        elif isdir(results_dir):
            for file in index_files:
                file_path = join(results_dir, file)
                if not isfile(file_path):
                    raise ValueError(f"{file} is not found in results directory!")
            kmer_size, genomes, positions, mem_blocks = retrieve_metadata(join(results_dir, 'metadata.json'))
            self.number_of_genomes = len(genomes)
            self.scores = {i: tuple(self.decode_score(bytes(s))) for i, s in enumerate(load_scores(str(results_dir), ''))}
            self.kmers_stream = gzip.open(join(results_dir, 'kmers.bgz'), 'rb')
            self.indices_stream = gzip.open(join(results_dir, 'indices.bgz'), 'rb')
        else:
            raise ValueError(f"{results_dir} is not a valid directory!")
        self.results_dir = results_dir
        self.kmer_size = kmer_size
        self.genomes = genomes
        self.positions = positions
        self.kmer_bitsize = math.ceil(((kmer_size*2))/8)
        self.i_bitsize = 8
        self.threads = threads
        self.mem_blocks = mem_blocks
        self.kmer_buffer = io.BufferedReader(self.kmers_stream)
        self.i_buffer = io.BufferedReader(self.indices_stream)
        
    def __iter__(self) -> PKIterator:
        """ Object iterator

        Returns
        -------
        PKIterator
            returns a kmer and a score in bytes
        """        
        return PKIterator(self)

        
    def get_kmer_byte(self) -> bytes:
        """read bytes in size of one kmer from kmers file

        Returns
        -------
        bytes
            bytes representing Kmer
        """        
        kmer_byte = self.kmer_buffer.read(self.kmer_bitsize)
        if not kmer_byte:
            return None
        return kmer_byte
        
    def get_score_byte(self) -> bytes:
        """read bytes in size of one score from scores file

        Returns
        -------
        bytes
            bytes represting what samples have a Kmer
        """        
        i_byte = self.i_buffer.read(self.i_bitsize)
        if not i_byte:
            return None
        i = int.from_bytes(i_byte, byteorder='big', signed=False)
        return i_byte, self.scores[i]
    
    def seek_kmer(self, kmer_num : int):
        """Move Kmer file pointer past a number of Kmers

        Parameters
        ----------
        kmer_num : int
            Number of Kmers to skip
        """        
        self.kmer_buffer.seek(kmer_num*self.kmer_bitsize)
    
    def seek_i(self, i_num: int):
        """Move Score file pointer past a number of scores

        Parameters
        ----------
        score_num : int
            Number of scores to skip
        """        
        self.i_buffer.seek(i_num*self.i_bitsize)

    def reset_iter(self):
        """Move Kmer and Score files' pointers to 0 position
        """        
        self.seek_kmer(0)
        self.seek_i(0)
        
    def set_threads(self, threads : int):
        """Set number of threads to run multithreaded processes

        Parameters
        ----------
        threads : int
            Number of threads
        """        
        self.threads = threads
    
    def decode_score(self, bits : bytes) -> list:
        """Convert score bytes to a list of binary flags


        Parameters
        ----------
        bits : bytes
            Score bytes generated by PanKmer indexing

        Returns
        -------
        list
            List of binary flags indicating if genome at position x has the corresponding Kmer
            0: Genome doesn't have Kmer
            1: Genome has Kmer
        """        
        return score_byte_to_blist(bits, self.number_of_genomes)
        
#     def encode_score(self, present_samples):
        
    
    def decode_kmer(self, bits : bytes) -> int:
        """Convert Kmer bytes to an integer (C unsigned long long)

        Parameters
        ----------
        bits : bytes
            Bytes representing Kmer generated by PanKmer indexing

        Returns
        -------
        int
            Python integer represting the Kmer represented by bytes
        """        
        return int.from_bytes(bits, byteorder='big')
    
    def encode_kmer(self, kmer : str) -> int:
        """Generate a integer represting canonical Kmer

        Parameters
        ----------
        kmer : str
            Sequence representing one Kmer only

        Returns
        -------
        int
            Python integer representing canonical Kmer
        """        
        upper = (1<<(self.kmer_size*2))-1
        lower = 0
        kmer_byte = bytes(kmer.upper(), 'ascii')
        kmer_list = break_seq(kmer_byte, upper, lower)
        if len(kmer_list) > 1:
            raise ValueError(f"encode_kmer should be used for one Kmer of size {self.kmer_size} only! "
                             f"Please use get_sequence_kmerbits for a sequence larger than that.")
        return kmer_list[0]
    
    def get_sequence_kmerbits(self, seq : str, upper : int = None, lower : int = None) -> list:
        """Get canonical Kmers making up the sequence

        Parameters
        ----------
        seq : str
            Sequence to Kmerize

        Returns
        -------
        list
            list of Python integers representing the canonical Kmers
        """        
        if upper == None:
            upper = (1<<(self.kmer_size*2))-1
        if lower == None:
            lower = 0
        seq_byte = bytes(str(seq).upper(), 'ascii')
        kmer_bits = break_seq(seq_byte, upper, lower)
        
        return kmer_bits
    
    def get_kmer_scores(self, kmers : dict = None) -> dict:
        """Retrieve scores of Kmer in the Kmers dictionary.
        if no dictionary is supplied then retrieve all scores.

        Parameters
        ----------
        kmers : dict, optional
            Dictionary where the keys are target Kmers and values are default values
            to populate with if the Kmer isn't found in the index, by default None

        Returns
        -------
        dict
            Populated dictionary where keys are Kmers and values are scores
        """
        if kmers == None:
            # get all scores
            kmers = {}
            for kmer, score in self:
                kmers[kmer] = score
        else:
            # get only scores in kmers
            for kmer, score in self:
                kmer_int = self.decode_kmer(kmer)
                if kmer_int in kmers:
                    kmers[kmer_int] = score
        return kmers
    
    def get_sequence_score(self, seq : str, kmers : dict = None) -> list:
        """Retrieve scores of canonical Kmers in the sequence

        Parameters
        ----------
        seq : str
            Target sequence
        kmers : dict, optional
            dictionary of Kmer:Score to use to assign scores to Kmers in sequence.
            If no dictionary is supplied then the index will be used, by default None

        Returns
        -------
        list
            List of integers representing score of Kmer ending at postion x in the sequence
        """        
        kmer_bits = self.get_sequence_kmerbits(seq)
        if kmers == None:
            kmers = {kmer:None for kmer in kmer_bits}
            kmers = self.get_kmer_scores(kmers=kmers)
        return [kmers[kmer_bit] for kmer_bit in kmer_bits]

    def get_fasta_kmerbits(self, fasta_path : str, upper : int = None,
                           lower : int = None) -> set:
        """Get integers representing all canonical Kmers in a fasta file

        Parameters
        ----------
        fasta_path : str
            Path to fasta file
        upper : int, optional
            Python integer of the upper bound limit of canonical Kmer, by default None
        lower : int, optional
            Python integer of the lower bound limit of canonical Kmer, by default None

        Returns
        -------
        set
            A set of integers representing all canonical Kmers in fasta file
        """        
        kmer_set = {}
        with gzip_agnostic_open(fasta_path, 'rt') as infile:
            for record in SeqIO.parse(infile, 'fasta'):
                kmers = self.get_sequence_kmerbits(str(record.seq), upper, lower)
                kmer_set = kmer_set.union(set(kmers))
        return kmer_set
        
    def get_positional_fasta_kmerbits(self, fasta_path : str,
                                      contig_header : str = 'id') -> dict:
        """Get lists of integers representing each canonical Kmer in each position
        in each contig in the fasta file

        Parameters
        ----------
        fasta_path : str
            Path to fasta file
        contig_header : str, optional
            Header style to retrieve for contigs (id or description), by default 'id'

        Returns
        -------
        dict
            A dictionary where keys are contigs' names and values are lists of
            integers representing each canonical Kmer in each position in the corresponding contig
        """        
        upper = (1<<(kmer_size*2))-1
        lower = 0
        return get_positional_kmers(fasta_path, upper, lower, contig_header)

    def get_sequences_scores(self, sequences: dict, cpu: int = None) -> dict:
        """
        Get per-position scores for a dictionary of sequences

        Parameters
        ----------
        sequences : dict
            A dictionary of sequences. Key = name, value = sequence
        cpu : int, optional
            Number of cores to use. This will replicate memory used by parent function.

        Returns
        -------
        dict
            A dictionary of scores. Key = name, value = list of per-position scores
        """
        if cpu == None:
            cpu = self.threads
        # Get sequences kmers and sort them
        sequences_kmers = {name:self.get_sequence_kmerbits(seq) for name, seq in sequences.items()}
        combined_kmers = set()
        scored_kmers = {}
        for name, kmer_bits in sequences_kmers.items():
            for kmer_bit in kmer_bits:
                combined_kmers.add(kmer_bit)
        combined_kmers = sorted(list(combined_kmers))
        # divide kmers into buckets of kmers
        chunk_size = math.ceil(len(combined_kmers)/cpu)
        core_chunks = [[self.results_dir, {kmer:0 for kmer in combined_kmers[i:i+chunk_size]}] for i in range(0, len(combined_kmers), chunk_size)]
        # Update the kmers with their scores from the index
        if cpu == 1:
            results = map(get_sorted_kmer_scores, core_chunks)
        else:
            core_worker = Pool(cpu)
            results = core_worker.map(get_sorted_kmer_scores, core_chunks)
        for result in results:
            scored_kmers.update(result)
        
        return {name:self.get_sequence_score(seq, scored_kmers) for name, seq in sequences.items()}
    
    def get_regional_scores(self, reference: str, regions: dict,
                            cpu: int = None) -> dict:
        """
        Retrieve regions from the reference and return their per-position Kmer scores

        Parameters
        ----------
        reference : str
            Path to a reference file in GZIP fasta format.
        regions : dict
            Dictionary of regions. Key = name of contig, values = list of start and end positions to extract.
            If no start, end positions are given, the function will use the entire contig.
            example = {'contig_1': [[4, 10], [59, 90]], 'contig_2': []}
        cpu : int, optional
            Number of cores to use. This will replicate memory used by parent function., default to PKResults threads.

        Returns
        -------
        dict
            A nested dictionary of scores. Key = contig, value = dictionary where key = tuple of start and end positons
            and value = list of per-position scores.
            example: {'contig_1': {(4, 10): [scores], (59, 90): [scores]}}
        """
        if cpu == None:
            cpu = self.threads
        sequences = {}
        with gzip_agnostic_open(reference, 'rt') as infile:
            for record in SeqIO.parse(infile, 'fasta'):
                # sequences[record.id] = {}
                if record.id in regions:
                    if len(regions[record.id]) > 0:
                        for start, end in regions[record.id]:
                            seq  = str(record.seq)[start:end]
                            sequences[(record.id, start, end)] = seq
                    else:
                        seq = str(record.seq)
                        sequences[(record.id, 0, len(seq))] = seq
        sequence_scores = self.get_sequences_scores(sequences=sequences, cpu=cpu)
        sequences_reduced = {}
        for id, start, end in sequences.keys():
            if id not in sequences_reduced:
                sequences_reduced[id] = {}
            sequences_reduced[id][(start, end)] = sequence_scores[(id, start, end)]


        return sequences_reduced
    
    def get_collapsed_regional_scores(self, reference : str, regions : dict) -> dict:
        """
        Retrieve regions from the reference and return their per-position Kmer scores

        Parameters
        ----------
        reference : int
            Path to a reference file in GZIP fasta format.
        regions : int
            Dictionary of regions. Key = name of contig, values = list of start and end positions to extract.
            If no start, end positions are given, the function will use the entire contig.
            example = {'contig_1': [[4, 10], [59, 90]], 'contig_2': []}

        Returns
        -------
        dict
            _description_
        """        
        regional_scores = self.get_regional_scores(reference, regions)
        blevel_scores = {}
        for id, regions in regional_scores.items():
            blevel_scores[id] = {}
            for region, scores in regions.items():
                blevel_scores[id][region] = self.get_blevel_scores(scores)

        return blevel_scores

    def get_blevel_scores(self, scores: list) -> list:
        '''
        Get base level score for each position in the reference
        '''
        
        fillna_val = [0] * self.number_of_genomes
        scores = [scores[0] for i in range(self.kmer_size-1)] + scores
        for c, s in enumerate(scores[self.kmer_size:], start=self.kmer_size):
            for i in range(self.kmer_size): 
                if c-i < 0:
                    break
                scores[c-i] = tuple(max(b1, b2) for b1, b2 in zip((scores[c-i] or fillna_val), (s or fillna_val)))
                # When index does not contain ref, scores[c-i] or s may be NoneType, which causes an error
        return [(s or fillna_val) for s in scores]
    
    def get_regional_coverage(self, reference : str, regions : dict) -> dict:
        '''
        Get per position coverage for each region in the reference specified by the 
        regions dictionary.

        Parameters
        ----------
        reference: str, required
            A path to the reference file in fasta format
        regions: dict, required
            A dictionary whose keys are contigs in the refernce and the values
            are lists of start and end positons

        Returns
        -------
        A dictionary whose keys are contig names and start and end positions and 
        values are lists of number of samples in the index that have that position
        '''
        coverage = {}
        regional_scores = self.get_collapsed_regional_scores(reference, regions)
        for id, regions in regional_scores.items():
            coverage[id] = {}
            for region, scores in regions.items():
                coverage[id][region] = list(np.array(scores).sum(axis=1))
        
        return coverage

    def close(self):
        self.kmers_stream.close()
        self.scores_stream.close()
        if self.input_is_tar:
            self.tar.close()
        

# def update_genomes(index, genomes, genomes_dict):
#     results = PKResults(index)
#     index_genomes = results.genomes
#     genomes_size_dict = {}
#     for genome, genome_size in genomes_dict.items():
#         genomes_size_dict[genome] = genome_size
#     for genome in index_genomes:
#         genomes_size_dict[genome] = index_genomes[genome]
    
#     genomes = genomes + [i for i in index_genomes]
#     return genomes, genomes_size_dict


# def update_index(upper, lower, kmer_bitsize, score_bitsize, genomes, outdir, index_dir):
#     gnum = len(genomes)
#     kmers_post = {}
#     kmers = get_kmers(upper, lower, genomes)
#     print_err(f"Saving {lower}-{upper} kmers.")
#     kmers_out_path = join(outdir, f'kmers_{lower}_{upper}.b.gz')
#     scores_out_path = join(outdir, f'scores_{lower}_{upper}.b.gz')
#     kmers_exist = sorted(kmers.keys())

#     results = PKResults(index_dir)
#     total_genomes_number = score_bitsize + results.number_of_genomes
#     new_score_bitsize = math.ceil(total_genomes_number/8)
#     # Find lower bound position from index
#     initial_lower_bound = get_lower_bound(results.positions, lower)
#     # Move iterator from initial lower bound to actual lower bound
#     if initial_lower_bound != None:
#         results.seek_kmer(initial_lower_bound)
#         results.seek_score(initial_lower_bound)
#     lb_bit = int(lower).to_bytes(results.kmer_bitsize, byteorder="big", signed=False) # lower bound bit
#     ub_bit = int(upper).to_bytes(results.kmer_bitsize, byteorder="big", signed=False) # upper bound bit
#     results_iter = iter(results)
#     try:
#         index_kmer, index_score = next(results_iter)
#     except StopIteration:
#         index_kmer, index_score = [None, None]
#     while index_kmer != None and index_kmer < lb_bit:
#         try:
#             index_kmer, index_score = next(results_iter)
#         except StopIteration:
#             index_kmer, index_score = [None, None]    


#     dict_iter = iter(kmers_exist)
#     if kmers_exist:
#         dict_kmer = next(dict_iter)
#         dict_kmer_bit = dict_kmer.to_bytes(results.kmer_bitsize,
#                                         byteorder="big", signed=False)
#     else:
#         dict_kmer = None
#     count = 0
#     kmer_to_write = None
#     score_to_write = None
#     with gzip.open(kmers_out_path, 'wb') as kmers_out, gzip.open(scores_out_path,'wb') as scores_out:
#         with io.BufferedWriter(scores_out, buffer_size=1000*score_bitsize) as so_buffer ,\
#             io.BufferedWriter(kmers_out, buffer_size=1000*kmer_bitsize) as ko_buffer:
#             while (index_kmer != None and index_kmer <= ub_bit) or dict_kmer != None:
#                 if count%10000000 == 0 and count != 0:
#                     kmers_post[kmer_to_write] = count
#                     count = 0
#                 # Current samples don't have any Kmers left
#                 # but pre-existing index still has Kmers
#                 if dict_kmer == None and index_kmer != None:
#                     score = int.from_bytes(index_score, 'big', signed=False)
#                     index_kmer_int = int.from_bytes(index_kmer, 'big', signed=False)
#                     kmer_to_write = index_kmer_int
#                     score_to_write = score << gnum
#                     try:
#                         index_kmer, index_score = next(results_iter)
#                     except StopIteration:
#                         index_kmer, index_score = [None, None]
#                 # Current samples have Kmers left
#                 # but pre-existing index doesn't have any Kmers left
#                 elif dict_kmer != None and index_kmer == None:
#                     kmer_to_write = dict_kmer
#                     score_to_write = kmers[dict_kmer]
#                     try:
#                         dict_kmer = next(dict_iter)
#                         dict_kmer_bit = dict_kmer.to_bytes(results.kmer_bitsize,
#                                                         byteorder="big", signed=False)
#                     except StopIteration:
#                         dict_kmer = None
#                 # Current sample Kmer and pre-existing Kmer is the same
#                 elif index_kmer == dict_kmer_bit:
#                     score = int.from_bytes(index_score, 'big', signed=False)
#                     new_score = (score << gnum) | kmers[dict_kmer]
#                     # kmers[dict_kmer_bit] = new_score
#                     kmer_to_write = dict_kmer
#                     score_to_write = new_score
#                     try:
#                         dict_kmer = next(dict_iter)            
#                         dict_kmer_bit = dict_kmer.to_bytes(results.kmer_bitsize,
#                                                     byteorder="big", signed=False)
#                     except StopIteration:
#                         dict_kmer = None

#                     try:
#                         index_kmer, index_score = next(results_iter)
#                     except StopIteration:
#                         index_kmer, index_score = [None, None]
#                 # Add pre-existing Kmer
#                 elif index_kmer < dict_kmer_bit or dict_kmer == None:
#                     score = int.from_bytes(index_score, 'big', signed=False)
#                     index_kmer_int = int.from_bytes(index_kmer, 'big', signed=False)
#                     new_score = score << gnum
#                     # kmers[index_kmer] = new_score
#                     kmer_to_write = index_kmer_int
#                     score_to_write = new_score

#                     try:
#                         index_kmer, index_score = next(results_iter)
#                     except StopIteration:
#                         index_kmer, index_score = [None, None]
#                 # Add current sample Kmer
#                 else:
#                     kmer_to_write = dict_kmer
#                     score_to_write = kmers[dict_kmer]
#                     try:
#                         dict_kmer = next(dict_iter)
#                         dict_kmer_bit = dict_kmer.to_bytes(results.kmer_bitsize,
#                                                         byteorder="big", signed=False)
#                     except StopIteration:
#                         dict_kmer = None
#                 a = ko_buffer.write(
#                     kmer_to_write.to_bytes(kmer_bitsize,
#                     byteorder="big", signed=False))
#                 b = so_buffer.write(
#                     score_to_write.to_bytes(new_score_bitsize,
#                     byteorder="big", signed=False))
#                 count += 1
#             if kmer_to_write != None and kmer_to_write not in kmers_post:
#                 kmers_post[kmer_to_write] = count-1

#     return kmers_post


# def count_scores(scores_file, score_bitsize):
#     scores_counts = {}
#     # Read kmers and scores files in byte mode
#     with gzip.open(scores_file, "rb") as score_in:
#         # Buffer file reading
#         score_buffer = io.BufferedReader(score_in)
#         # Read score/kmer sized bytes from the files
#         score = score_buffer.read(score_bitsize)
#         while score:
#             if score in scores_counts:
#                 scores_counts[score]+=1
#             else:
#                 scores_counts[score] = 1
#             score = score_buffer.read(score_bitsize)

#     return scores_counts


def get_kmers_wrapper(seq, upper, lower):
    seq_byte = bytes(str(seq).upper(), 'ascii')
    kmers = break_seq(seq_byte, upper, lower)
    return kmers


def index_wrapper(genomes_input, outdir: str, split_memory: int = 1,
                  threads: int = 1, index: str = '', fraction: float = 1.0,
                  k=31):
    if basename(outdir).endswith('.tar'):
        output_is_tar = True
        outdir = join(dirname(outdir), basename(outdir)[:-4])
    else:
        output_is_tar = False

    if not isinstance(genomes_input, (tuple, list)):
        genomes_input = list(genomes_input) if isinstance(genomes_input, set) else [genomes_input]
    if len(genomes_input) == 1 and isfile(genomes_input[0]) and tarfile.is_tarfile(genomes_input[0]):
        input_is_tar = True
        with tarfile.open(genomes_input[0]) as tar:
            genomes = [tarinfo.name for tarinfo in tar if tarinfo.isreg()]
    else:
        input_is_tar = False
        genomes = list(itertools.chain.from_iterable(
            ([join(g, f) for f in sorted(listdir(g))] if isdir(g) else [g])
            for g in genomes_input))
        # Check if input files exist and are files
        for genome in genomes:
            if not exists(genome) or not isfile(genome):
                raise RuntimeError(f"{genome} does not exist or is not a file!")
                return 1

    # Make the output directory if it doesn't exist
    if not exists(outdir):
        mkdir(outdir)

    print_err('Recording genome sizes')
    genomes_dict = {}
    for genome in genomes:
        print_err(f'Measuring {genome}')
        genomes_dict[genome] = 0
        if input_is_tar:
            tar = tarfile.open(genomes_input[0])
            infile = gzip_agnostic_open(genome, 'rt', tar=tar)
        else:
            infile = gzip_agnostic_open(genome, 'rt')
        for record in SeqIO.parse(infile, 'fasta'):
            genomes_dict[genome] += len(str(record.seq))
        infile.close()
        if input_is_tar:
            tar.close()

    positions_dictionary, mem_blocks = run_index((str(genomes_input[0]) if input_is_tar else ''),
            genomes, str(outdir),
            split_memory=split_memory, threads=threads,
            index=index, input_is_tar=input_is_tar, fraction=fraction,
            kmersize=kmer_size())
    
    # if index:
    #     genomes, genomes_dict = update_genomes(index, genomes, genomes_dict)

    print_err("Saving metadata.")
    metadata_dict = {'kmer_size': kmer_size(),
        'version': __version__,
        'genomes': {c:g for c, g in enumerate(genomes)},
        'genome_sizes': genomes_dict,
        'positions': dict(sorted(positions_dictionary.items(), key=itemgetter(1))),
        'mem_blocks': mem_blocks
    }
    with open(f'{outdir}/metadata.json', 'w') as outfile:
        json.dump(metadata_dict, outfile)
    if output_is_tar:
        with tarfile.open(f'{outdir}.tar', 'w') as tar:
            tar.add(outdir)
        shutil.rmtree(outdir)

def run_index_wrapper(args):
    if args.time:
        start = time.time()
    # index = args.index
    # index_wrapper(genomes_input, fraction, 31, output)
    index_wrapper(args.genomes, args.output, split_memory=args.split_memory,
                  threads=args.threads, index='', k=31,
                  fraction=args.fraction)
    if args.time:
        stop = time.time()
        print(f'Indexed in {(stop - start) / 60:.2f} minutes')


def run_count(args):
    kmer_counts = count_kmers(*(PKResults(i) for i in args.index), names=args.index)
    kmer_counts.to_csv(sys.stdout, sep='\t')


def run_collect(args):
    sat_df = collect(PKResults(args.index), args.output, title=args.title,
                        width=args.width, height=args.height,
                        palette=args.color_palette, alpha=args.alpha,
                        linewidth=args.linewidth, conf=args.conf,
                        contours=args.contours, legend_loc=args.legend_loc)
    if args.table:
        sat_df.to_csv(args.table, sep='\t', index=False)


def run_upset(args):
    if args.time:
        start = time.time()
    upset(PKResults(args.input), args.output, args.genomes,
          vertical=args.vertical,
          show_counts=args.show_counts,
          min_subset_size=args.min_subset_size,
          max_subset_size=args.max_subset_size,
          exclusive=args.exclusive,
          table=args.table)
    if args.time:
        stop = time.time()
        print(f'UpSet plot generated in {(stop - start) / 60:.2f} minutes')


def run_subset(args):
    if args.time:
        start = time.time()
    subset(PKResults(args.input), args.output, args.genomes,
          exclusive=args.exclusive)
    if args.time:
        stop = time.time()
        print(f'subset index generated in {(stop - start) / 60:.2f} minutes')


def run_adjacency_matrix(args):
    if args.time:
        start = time.time()
    results = PKResults(args.input_dir)
    output = args.output
    df = get_adjacency_matrix(results)
    df.to_csv(output)
    if args.time:
        stop = time.time()
        print(f'Matrix generated in {(stop - start) / 60:.2f} minutes')


def run_tree(args):
    adj_matrix = pd.read_csv(args.input, index_col=0)
    tree(adj_matrix, newick=args.newick, metric=args.metric, method=args.method,
         transformed_matrix=args.transformed_matrix)


def run_clustermap(args):
    adj_matrix = pd.read_csv(args.input, index_col=0)
    clustermap(adj_matrix, args.output, cmap=args.colormap, width=args.width,
        height=args.height, metric=args.metric, method=args.method,
        heatmap_tick_pos=args.heatmap_ticks,
        cbar_tick_pos=args.cbar_ticks, dendrogram_ratio=args.dend_ratio,
        dendrogram_spacer=args.dend_spacer)



def run_regcoverage(args):
    if args.output:
        reg_coverage(*(PKResults(i) for i in args.index), ref=args.ref,
            coords=args.coords, output_file=args.output, bgzip=args.bgzip,
            genes=args.genes, flank=args.flank, processes=args.processes)
    else:
        for chrom, start, end, *values in reg_coverage(
            *(PKResults(i) for i in args.index), ref=args.ref,
                coords=args.coords, genes=args.genes, flank=args.flank,
                processes=args.processes):
            print(chrom, start, end, *values, sep='\t')


def run_genome_coverage(args):
    genome_coverage(*(PKResults(i) for i in args.index), output=args.output,
        ref=args.ref, chromosomes=args.chromosomes, output_table=args.table,
        groups=args.groups, title=args.title, x_label=args.x_label,
        legend=args.legend,
        legend_title=args.legend_title, legend_loc=args.legend_loc,
        bin_size=args.bin_size, width=args.width, height=args.height,
        color_palette=args.color_palette, alpha=args.alpha,
        linewidth=args.linewidth, processes=args.processes)


def run_genome_coverage_plot(args):
    plotting_data=pd.read_table(args.table)
    x_label = args.x_label if args.x_label else plotting_data.columns[1]
    legend_title = args.legend_title if args.legend_title else plotting_data.columns[3]
    columns = (
        plotting_data.columns[0],
        x_label,
        plotting_data.columns[2],
        legend_title,
        f'{legend_title}_chrom'
    )
    plotting_data.columns = columns
    if args.chromsizes:
        sizes = pd.read_table(args.chromsizes, header=None)
        sizes.columns = 'name', 'size'
    else:
        sizes = None
    genome_coverage_plot(plotting_data, output=args.output,
        groups=args.groups, loci=args.loci, sizes=sizes,
        title=args.title, x_label=x_label, legend=args.legend,
        legend_title=legend_title,
        legend_loc=args.legend_loc, width=args.width, height=args.height,
        color_palette=args.color_palette, alpha=args.alpha,
        linewidth=args.linewidth)


def run_coverage_heatmap(args):
    coverage_heatmap(PKResults(args.index), args.refs, args.features,
        args.output, n_features=args.n_features, width=args.width,
        height=args.height)


def run_dryrun(args):
    metadata_dict = dryrun(args.genomes, split_memory=args.split_memory, threads=args.threads)
    json.dump(metadata_dict, sys.stdout)


def run_download_example(args):
    download_example(args.dir, args.bacterial, args.n_samples)


def main():
    parser = argparse.ArgumentParser(
        add_help=True)
    parser.add_argument('--version', action='version',
                    version='%(prog)s {version}'.format(version=__version__))
    subparsers = parser.add_subparsers(dest='func')
    index_parser = subparsers.add_parser('index',
        help='generate k-mer index')
    index_parser.add_argument(
        '-g', '--genomes', metavar='<genomes[.tar]>',
        required=True,
        nargs='+',
        help='directory containig input genomes')
    index_parser.add_argument(
        '-o', '--output', metavar='<output[.tar]>',
        required=True,
        help='output directory or tarfile that will contain the k-mer index')
    index_parser.add_argument(
        '--split-memory', metavar='<int>',
        type=int, help=(
        'Parallel.'
        'This splits the indexing into multiple rounds; reducing memory.'
        ), default=1)
    index_parser.add_argument(
        '-t', '--threads', metavar='<int>',
        type=int, help=('Number of threads to use'), default=1)
    index_parser.add_argument(
        '--fraction', metavar='<float>', type=float, default=1.0, help=(
        'Fraction of k-mer minimizers to use'))
    # index_parser.add_argument(
    #     '-i', '--index', metavar='<index[.tar]>',
    #     dest='index', required=False, default='',
    #     help=('Path to existing index to update.')
    # )
    index_parser.add_argument(
        '--time', action='store_true',
        help='Report the time required to execute'
    )
    index_parser.set_defaults(func=run_index_wrapper)

    count_parser = subparsers.add_parser('count',
        help='count k-mers in one or more indexes')
    count_parser.add_argument(
        '-i', '--index', metavar='<index[.tar]>', type=str, action='store',
        dest='index', required=True, nargs='+', help='a k-mer index')
    count_parser.set_defaults(func=run_count)

    collect_parser = subparsers.add_parser('collect',
        help='calculate k-mer collection curve')
    collect_parser.add_argument(
        '-i', '--index', metavar='<index[.tar]>', type=str, action='store',
        required=True, help='a k-mer index')
    collect_parser.add_argument(
        '-o', '--output', metavar='<output.{pdf,png,svg}>', type=str,
        action='store', dest='output',
        help='destination file for plot')
    collect_parser.add_argument(
        '-t', '--table', metavar='<output-table.tsv>',
        help='output TSV file containing plotted data')
    collect_parser.add_argument('--title', metavar='<"Plot title">',
        help='set the title for the plot')
    collect_parser.add_argument('--width', metavar='<float>', type=float,
        default=4.0, help='set width of figure in inches [4]')
    collect_parser.add_argument('--height', metavar='<float>', type=float,
        default=3.0, help='set height of figure in inches [3]')
    collect_parser.add_argument('--color-palette', metavar='<#color>', nargs='+',
        default=COL_COLOR_PALETTE, help='color palette to use')
    collect_parser.add_argument('--alpha', metavar='<float>', type=float,
        default=1.0, help='transparency value for lines [1.0]')
    collect_parser.add_argument('--linewidth', metavar='<int>', type=int,
        default=3, help='line width for plot [3]')
    conf_or_cont = collect_parser.add_mutually_exclusive_group()
    conf_or_cont.add_argument('--conf', action='store_true',
        help='calculate confidence intervals for collection curves')
    conf_or_cont.add_argument('--contours', metavar='<int>', type=int,
        nargs='+', help='set contours for collection curves (in percent)')
    collect_parser.add_argument('--legend-loc',
        choices=('best', 'upper left', 'upper right', 'lower left',
                 'lower right', 'outside'),
        default='best', help='location of legend [best]')
    collect_parser.set_defaults(func=run_collect)

    upset_parser = subparsers.add_parser('upset', help='generate upset plot')
    upset_parser.add_argument(
        '-i', '--input', metavar='<index[.tar]>',
        required=True, help='a k-mer index')
    upset_parser.add_argument(
        '-o', '--output', metavar='<output.{pdf,png,svg}>', required=True,
        help='destination file for upset plot')
    upset_parser.add_argument(
        '-g', '--genomes', metavar='<genome>', required=True, nargs='+',
        help='list of genomes to include')
    upset_parser.add_argument(
        '-v', '--vertical', action='store_true',
        help='draw the plot vertically')
    upset_parser.add_argument(
        '-x', '--exclusive', action='store_true',
        help='exclude k-mers that occur in genomes other than the input set')
    upset_parser.add_argument(
        '-t', '--table', metavar='<table.tsv[.gz]>',
        help='write (optionally compressed) table of values in tsv format')
    upset_parser.add_argument(
        '--show-counts', action='store_true',
        help='show counts for each subset')
    upset_parser.add_argument(
        '--min-subset-size', metavar='<int>', type=int,
        help='show only subsets larger than a minimum size')
    upset_parser.add_argument(
        '--max-subset-size', metavar='<int>', type=int,
        help='show only subsets smaller than a maximum size')
    upset_parser.add_argument(
        '--time', action='store_true',
        help='report the time required to execute')
    upset_parser.set_defaults(func=run_upset)

    subset_parser = subparsers.add_parser('subset', help='subset an index')
    subset_parser.add_argument(
        '-i', '--input', metavar='<input-index[.tar]>',
        required=True, help='a k-mer index')
    subset_parser.add_argument(
        '-o', '--output', metavar='<output-index[.tar]>', required=True,
        help='destination file for subset index')
    subset_parser.add_argument(
        '-g', '--genomes', metavar='<genome>', required=True, nargs='+',
        help='list of genomes to include '
    )
    subset_parser.add_argument(
        '-x', '--exclusive', action='store_true',
        help='exclude k-mers that occur in genomes other than the input set'
    )
    subset_parser.add_argument(
        '--time', action='store_true',
        help='report the time required to execute'
    )
    subset_parser.set_defaults(func=run_subset)

    adjmat_parser = subparsers.add_parser('adj-matrix',
        help='generate adjacency matrix')
    adjmat_parser.add_argument(
        '-i', '--input', metavar='<index[.tar]>',
        dest='input_dir', required=True, help='a k-mer index')
    adjmat_parser.add_argument(
        '-o', '--output', metavar='<adjmatrix.csv>',
        dest='output', required=True,
        help='destination file for adjacency matrix')
    adjmat_parser.add_argument(
        '--time', action='store_true',
        help='Report the time required to execute')
    adjmat_parser.set_defaults(func=run_adjacency_matrix)

    tree_parser = subparsers.add_parser('tree',
        help='generate a heirarchical clustering tree from an adjacency matrix')
    tree_parser.add_argument(
        '-i', '--input', metavar='<adjmatrix.csv>',
        dest='input', required=True, help='adjacency matrix file')
    tree_parser.add_argument(
        '-n', '--newick', action='store_true', dest='newick',
        help='output tree in NEWICK format'
    )
    tree_parser.add_argument(
        '--metric', choices=('intersection', 'jaccard', 'overlap'),
        type=str, action='store', dest='metric', default='intersection',
        help='similarity metric [intersection]')
    tree_parser.add_argument(
        '--method', choices=('single', 'complete', 'average', 'weighted', 'centroid'),
        type=str, action='store', dest='method', default='complete',
        help='clustering method [complete]')
    tree_parser.add_argument(
        '--transformed-matrix', metavar='<matrix.csv>',
        help='Write similarity transformed matrix to file')
    tree_parser.set_defaults(func=run_tree)

    clustermap_parser = subparsers.add_parser('clustermap',
        help='plot a clustered heatmap from the adjacency matrix')
    clustermap_parser.add_argument(
        '-i', '--input', metavar='<adjmatrix.csv>', type=str, action='store',
        dest='input', required=True,
        help='adjacency matrix file')
    clustermap_parser.add_argument(
        '-o', '--output', metavar='<adjmatrix.{pdf,png,svg}>', type=str, action='store',
        dest='output', required=True,
        help='destination file for plot')
    clustermap_parser.add_argument(
        '--metric', choices=('intersection', 'jaccard', 'overlap'),
        type=str, action='store', dest='metric', default='intersection',
        help='similarity metric [intersection]')
    clustermap_parser.add_argument(
        '--method', choices=('single', 'complete', 'average', 'weighted', 'centroid'),
        type=str, action='store', dest='method', default='complete',
        help='clustering method [complete]')
    clustermap_parser.add_argument(
        '--colormap', metavar='<color_map>', type=str, action='store',
        default='mako_r', help='seaborn colormap for plot [mako_r]')
    clustermap_parser.add_argument(
        '--width', metavar='<float>', type=float, action='store',
        default=7.0, help='width of plot in inches [7]')
    clustermap_parser.add_argument(
        '--height', metavar='<float>', type=float, action='store',
        default=7.0, help='height of plot in inches [7]')
    clustermap_parser.set_defaults(func=run_clustermap)
    clustermap_parser.add_argument(
        '--heatmap-ticks', choices=('left', 'right'),
        default='left',
        help='Position of heatmap ticks. Must be "left" or "right" [left]')
    clustermap_parser.add_argument(
        '--cbar-ticks', choices=('left', 'right'),
        default='left',
        help='Position of color bar ticks. Must be "left" or "right" [left]')
    clustermap_parser.add_argument(
        '--dend-ratio', metavar='<float>', type=float, default=0.2,
        help='Fraction of plot width used for dendrogram [0.2]')
    clustermap_parser.add_argument(
        '--dend-spacer', metavar='<float>', type=float, default=0.1,
        help='Fraction of plot width used as spacer between dendrogram and heatmap [0.1]')
    clustermap_parser.set_defaults(func=run_clustermap)

    regcov_parser = subparsers.add_parser('reg-coverage',
        help='generate regional coverage')
    regcov_parser.add_argument(
        '-i', '--index', metavar='<index[.tar]>', type=str, action='store',
        dest='index', required=True, nargs='+', help='index')
    regcov_parser.add_argument(
        '-r', '--reference', metavar='<reference.fa.gz>', type=str, action='store',
        dest='ref', required=True, help='reference')
    regcov_parser.add_argument(
        '-c', '--coords', metavar='<chr:start-end>', type=str, action='store',
        dest='coords', required=True, help='genomic coordinates')
    regcov_parser.add_argument(
        '-o', '--output', metavar='<output.bdg[.gz]>', type=str,
        dest='output', help='write to file instead of standard output')
    regcov_parser.add_argument(
        '-b', '--bgzip', action='store_true',
        dest='bgzip', help='block compress the output file')
    regcov_parser.add_argument(
        '-g', '--genes', metavar='<genes.gff3[.gz]>', type=str,
        dest='genes', help='gff file of gene coordinates')
    regcov_parser.set_defaults(func=run_regcoverage)
    regcov_parser.add_argument(
        '-f', '--flank', metavar='<int>', type=int, default=0,
        dest='flank', help='size of flanking regions')
    regcov_parser.add_argument(
        '-p', '--processes', metavar='<int>', type=int, default=1,
        dest='processes', help='number of processes to use')
    regcov_parser.set_defaults(func=run_regcoverage)

    genomecov_parser = subparsers.add_parser('genome-coverage',
        help='generate genomic coverage')
    genomecov_parser.add_argument(
        '-i', '--index', metavar='<index[.tar]>', type=str, action='store',
        dest='index', required=True, nargs='+', help='index')
    genomecov_parser.add_argument(
        '-o', '--output', metavar='<output.{pdf,png,svg}>', type=str,
        action='store', dest='output', required=True,
        help='destination file for plot')
    genomecov_parser.add_argument(
        '-r', '--reference', metavar='<reference.fa.gz>', type=str,
        action='store', dest='ref', required=True,
        help='reference in BGZIP compressed FASTA format')
    genomecov_parser.add_argument(
        '-c', '--chromosomes', metavar='<chrX>', type=str, action='store',
        dest='chromosomes', required=True, nargs='+',
        help='chromosomes to include')
    genomecov_parser.add_argument(
        '-t', '--table', metavar='<output-table.tsv>',
        help='output TSV file containing plotted data')
    genomecov_parser.add_argument('--groups', metavar='<"Group">', nargs='+',
        help='list of groups for provided indexes [0]')
    genomecov_parser.add_argument('--title', metavar='<"Plot title">',
        default='Coverage', help='set the title for the plot')
    genomecov_parser.add_argument('--x-label', metavar='<"Label">',
        default='Chromosome', help='set the x-axis label for the plot')
    genomecov_parser.add_argument('--legend', action='store_true',
        help='include a legend with the plot')
    genomecov_parser.add_argument('--legend-title', metavar='<"Title">',
        default='Group', help='title of legend')
    genomecov_parser.add_argument('--legend-loc',
        choices=('best', 'upper left', 'upper right', 'lower left',
                 'lower right', 'outside'),
        default='best', help='location of legend [best]')
    genomecov_parser.add_argument('--bin-size', metavar='<int>', type=int, default=0,
        choices=(-2,-1,0,1,2),
        help=('Set bin size. The input <int> is converted to the bin size by '
              'the formula: 10^(<int>+6) bp. The default value is 0, i.e. '
              '1-megabase bins. [0]'))
    genomecov_parser.add_argument('--width', metavar='<float>', type=float,
        default=7.0, help='set width of figure in inches [7]')
    genomecov_parser.add_argument('--height', metavar='<float>', type=float,
        default=3.0, help='set height of figure in inches [3]')
    genomecov_parser.add_argument('--color-palette', metavar='<#color>', nargs='+',
        default=COLOR_PALETTE, help='color palette to use')
    genomecov_parser.add_argument('--alpha', metavar='<float>', type=float,
        default=0.5, help='transparency value for lines [0.5]')
    genomecov_parser.add_argument('--linewidth', metavar='<int>', type=int,
        default=3, help='line width for plot [3]')
    genomecov_parser.add_argument('--processes', metavar='<int>', type=int,
        default=1, help='number of processes to use')
    genomecov_parser.set_defaults(func=run_genome_coverage)

    gcplot_parser = subparsers.add_parser('gcplot',
        help='generate a plot from genomic coverage results')
    gcplot_parser.add_argument(
        '-t', '--table', metavar='<genomecov.tsv>', type=str, action='store',
        dest='table', required=True, help='genomic coverage results')
    gcplot_parser.add_argument(
        '-o', '--output', metavar='<output.{pdf,png,svg}>', type=str,
        action='store', dest='output', required=True,
        help='destination file for plot')
    gcplot_parser.add_argument('--groups', metavar='<"Group">', nargs='+',
        help='list of groups for provided indexes')
    gcplot_parser.add_argument('--loci', metavar='<"chr:pos:name">', nargs='+',
        help='list of loci to mark on plot')
    gcplot_parser.add_argument('--chromsizes', metavar='<file.chrom.sizes>',
        help='chromsizes file of the reference used to generate the table')
    gcplot_parser.add_argument('--title', metavar='<"Plot title">',
        help='set the title for the plot')
    gcplot_parser.add_argument('--x-label', metavar='<"Label">',
        help='set x-axis label for the plot')
    gcplot_parser.add_argument('--legend', action='store_true',
        help='include a legend with the plot')
    gcplot_parser.add_argument('--legend-title', metavar='<"Title">',
        help='title of legend')
    gcplot_parser.add_argument('--legend-loc',
        choices=('best', 'upper left', 'upper right', 'lower left',
                 'lower right', 'outside'),
        default='best', help='location of legend [best]')
    gcplot_parser.add_argument('--width', metavar='<float>', type=float,
        default=7.0, help='set width of figure in inches [7]')
    gcplot_parser.add_argument('--height', metavar='<float>', type=float,
        default=3.0, help='set height of figure in inches [3]')
    gcplot_parser.add_argument('--color-palette', metavar='<#color>', nargs='+',
        default=COLOR_PALETTE, help='color palette to use')
    gcplot_parser.add_argument('--alpha', metavar='<float>', type=float,
        default=0.5, help='transparency value for lines [0.5]')
    gcplot_parser.add_argument('--linewidth', metavar='<int>', type=int,
        default=3, help='line width for plot [3]')
    gcplot_parser.set_defaults(func=run_genome_coverage_plot)

    covhm_parser = subparsers.add_parser('cov-heatmap',
        help='draw coverage heatmap')
    covhm_parser.add_argument(
        '-i', '--index', metavar='<index[.tar]>', type=str, action='store',
        dest='index', required=True, help='index')
    covhm_parser.add_argument(
        '-r', '--refs', metavar='<reference.fa.gz>', type=str, action='store',
        dest='refs', required=True, nargs='+', help='references')
    covhm_parser.add_argument(
        '-f', '--features', metavar='<features.gff3>', type=str, action='store',
        dest='features', required=True, nargs='+',
        help='GFF files defining features (such as genes)')
    covhm_parser.add_argument(
        '-o', '--output', metavar='<output.{pdf,png,svg}>', type=str, action='store',
        dest='output', required=True, help='output file (must end with ".png")')
    covhm_parser.add_argument(
        '--n-features', metavar='<int>', type=int, action='store',
        dest='n_features',
        help='use to limit plotting to the first n features for each genome')
    covhm_parser.add_argument(
        '--width', metavar='<float>', type=float, action='store',
        dest='width', default=7, help='width of plot in inches', )
    covhm_parser.add_argument(
        '--height', metavar='<float>', type=float, action='store',
        dest='height', default=7, help='height of plot in inches', )
    covhm_parser.set_defaults(func=run_coverage_heatmap)

    dryrun_parser = subparsers.add_parser('dryrun',
        help='perform a dry run of indexing and print metadata')
    dryrun_parser.add_argument(
        '-g', '--genomes', metavar='<genomes[.tar]>',
        required=True,
        nargs='+',
        help='directory containig input genomes')
    dryrun_parser.add_argument(
        '--split-memory', metavar='<int>',
        type=int, help=(
        'Parallel.'
        'This splits the indexing into multiple rounds; reducing memory.'
        ), default=1)
    dryrun_parser.add_argument(
        '-t', '--threads', metavar='<int>',
        type=int, help=('Number of threads to use'), default=1)
    dryrun_parser.set_defaults(func=run_dryrun)

    download_parser = subparsers.add_parser('download-example',
        help='download an example dataset')
    download_parser.add_argument(
        '-d', '--dir', metavar='<dir/>', type=str, action='store',
        dest='dir', default=EXAMPLE_DATA_DIR,
        help='destination directory for example data')
    download_parser.add_argument(
        '-b', '--bacterial', action='store_true', dest='bacterial',
        help='if True, download bacterial genomes')
    download_parser.add_argument(
        '-n', '--n-samples', metavar='<int>', type=int, default=1, action='store', dest='n_samples',
        help='number of bacterial samples to download, max 164 [1]')
    download_parser.set_defaults(func=run_download_example)

    args = parser.parse_args()
    args.func(args)



import urllib3
import shutil
import os
import os.path
import tarfile
import ftplib
from pankmer.env import (EXAMPLE_DATA_URL, EXAMPLE_DATA_DIR,
                         BACTERIAL_DATA_FTP, BACTERIAL_DATA_PATHS)

def download_example(dir: str = EXAMPLE_DATA_DIR, bacterial: bool = False,
                     n_samples: int = 1):
    """Download an example datatset

    Parameters
    ----------
    dir : str
        Destination directory for example data
    bacterial : bool
        If True, download bacterial genomes
    n_samples : int
        Number of bacterial samples to download, max 164 [1]
    """

    if n_samples > 164:
        raise RuntimeError('n_samples parameter must be <= 164')
    if bacterial:
        ftp = ftplib.FTP(BACTERIAL_DATA_FTP)
        ftp.login()
        for ftp_path in BACTERIAL_DATA_PATHS[:n_samples]:
            with open(os.path.join(dir, os.path.basename(ftp_path)), 'wb') as f:
                ftp.retrbinary(f'RETR {ftp_path}', f.write)
    else:
        http = urllib3.PoolManager()
        tar_file_path = os.path.join(dir, os.path.basename(EXAMPLE_DATA_URL))
        if os.path.exists(tar_file_path[:-7]):
            raise RuntimeError('destination already exists')
        with http.request('GET', EXAMPLE_DATA_URL, preload_content=False) as r, open(tar_file_path, 'wb') as out_file:
            shutil.copyfileobj(r, out_file)
        with tarfile.open(tar_file_path) as tar:
            tar.extractall(dir)
        os.remove(tar_file_path)
